"""Base classes for the dbispipeline."""
from abc import ABC
from abc import abstractmethod


class Loader(ABC):
    """Abstract base class of a dataloader."""

    @abstractmethod
    def load(self):
        """Returns the data loaded by the dataloader."""
        pass

    @property
    @abstractmethod
    def configuration(self):
        """Returns a dict-like representation of the configuration."""
        pass

    @property
    def run_count(self):
        """Returns how many configurations this dataloader will produce."""
        return 0

    @property
    def is_multiloader(self):
        """Returns if this dataloader will produce multiple configurations."""
        return self.run_count > 0


class TrainTestLoader(Loader):
    """Abstract dataloader with explicit train and test data."""

    @abstractmethod
    def load_train(self):
        """Returns the train data."""
        pass

    @abstractmethod
    def load_test(self):
        """Returns the test data."""
        pass

    def load(self):
        """Returns the data as a train, test tuple."""
        return self.load_train(), self.load_test()


class TrainValidateTestLoader(TrainTestLoader):
    """Abstract dataloader with explicit train, validation and test data."""

    @abstractmethod
    def load_validate(self):
        """Returns the validation data."""
        pass

    def load(self):
        """Returns the data as a train, validation, test tuple."""
        return self.load_train(), self.load_validate(), self.load_test()


class Evaluator(ABC):
    """Abstract base class of an evaluator."""

    @abstractmethod
    def evaluate(self, model, data):
        """
        Evaluates the pipeline based on the given dataset.

        Args:
            model: the model given in the pipeline.
            data: the data used for evaluation/training.

        Returns: An dict result.
        """
        pass

    @property
    @abstractmethod
    def configuration(self):
        """Returns a dict-like representation of this loader.

        This is for storing its state in the database.
        """
        pass

    def _check_loader_methods(self, loader, methods=None):

        if type(methods) == str:
            methods = [methods]

        for method in methods:
            m = getattr(loader, method, None)
            if not callable(m):
                raise ValueError(
                    f'The dataloader {loader} does not implement'
                    'the required method {m}. Required methods are: '
                    '{methods}')


class StorageHandler(ABC):
    """Abstract base class of a storage handler."""

    @abstractmethod
    def handle_result(self, result):
        """
        Handles the result of an evaluator.

        Args:
            result: passed from an evaluator.
        """
        pass



"""Contains loaders returning pandas DataFrames."""
import inspect

import pandas as pd

from dbispipeline.base import Loader
from dbispipeline.utils import LOGGER


class DataframeLoader(Loader):
    """A generic one-in-all solution."""

    def __init__(self, df_path, features=None, targets=None, filters=None,
                 **extra_args):
        """
        Loads the data from a file on disk.

        Args:
            df_path: the path to the dataset file. If the filename ends with
                ".pckl", a read_pickle is attempted. If it ends with ".csv",
                a read_csv is attempted. Otherwise, an Error is raised.
            features: Which columns the data part of the result should contain.
                if this argument is falsy, all columns are returned.
            targets: which columns should be returned as target.
                if this argument is falsy, no target columns are returend.
            filters: a list of functions that are called on the resulting
                dataframe before it is split in data and target.
            **extra_args: any extra arguments that are stored in the db-config
                to help identify this generic instance.
        """
        self.features = features
        self.targets = targets or []
        self.df_path = df_path
        self.filters = filters or []
        self.extra_args = extra_args

    def load(self):
        """Loads the data."""
        if self.df_path.endswith('.pckl'):
            df = pd.read_pickle(self.df_path)
        elif self.df_path.endswith('.csv'):
            df = pd.read_csv(self.df_path)
        else:
            raise ValueError('unknown format for dataset: %s' % self.df_path)
        if df.empty:
            raise ValueError('loaded empty dataframe!')

        for f in self.filters:
            filtered_df = f(df)
            if filtered_df.empty:
                LOGGER.info('before filtering: %s', df)
                raise ValueError('dataframe is empty after filtering with: '
                                 f'{inspect.getsource(f)}')
            else:
                df = filtered_df
        if not self.features:
            self.features = df.columns
        return df[self.features], df[self.targets]

    @property
    def configuration(self):
        """Returns a database representation of this loader."""
        return dict(
            features=self.features,
            targets=self.targets,
            df_path=self.df_path,
            extra_args=self.extra_args,
        )



"""Declaring the main method of this package."""
import argparse

from dbispipeline.cli.main import main

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='executes dbis config files')
    # TODO: configurationfile is required but it is i.e. not needed for restore
    parser.add_argument('plan', help='file that holds config')
    parser.add_argument('-f',
                        '--force',
                        action='store_true',
                        help='Run with dirty git')
    parser.add_argument('--dryrun',
                        action='store_true',
                        help='run without sending results to database.')
    parser.add_argument(
        '-m',
        '--mail',
        choices=[None, 'run', 'total'],
        default=None,
        help='Mail notification level. Choose one of [None, \'run\', \'total\''
        ']. If set no None, no mails will be sent. if set to \'run\', one info'
        ' mail will be sent for each run. If set to \'total\', one mail will '
        'be sent after the entire pipeline is complete.')
    parser.add_argument(
        '--restore',
        default=None,
        type=str,
        help='Restores the backup contained in the given file.')
    parser.add_argument(
        '--slurm',
        action='store_true',
        help='send this job to the slurm job queue instead of running it local'
    )
    parser.add_argument('-v',
                        '--verbose',
                        action='store_true',
                        help='increase logging')
    args = parser.parse_args()
    main(args.dryrun, args.force, args.verbose, args.slurm, args.restore,
         args.mail, args.plan)
